using Test
using Distributions
using StableRNGs
using StatsBase
using LinearAlgebra
using Random

using EnsembleKalmanProcesses.ParameterDistributions

@testset "ParameterDistributions" begin
    @testset "ParameterDistributionType" begin
        # Tests for the ParameterDistributionType
        d = Parameterized(Gamma(2.0, 0.8))
        @test d.distribution == Gamma(2.0, 0.8)

        d = Samples([1 2 3; 4 5 6])
        @test d.distribution_samples == [1 2 3; 4 5 6]
        d = Samples([1 4; 2 5; 3 6]; params_are_columns = false)
        @test d.distribution_samples == [1 2 3; 4 5 6]
        d = Samples([1, 2, 3, 4, 5, 6])
        @test d.distribution_samples == [1 2 3 4 5 6]
        d = Samples([1, 2, 3, 4, 5, 6]; params_are_columns = false)
        @test d.distribution_samples == reshape([1, 2, 3, 4, 5, 6], :, 1)

        # Vector container - 
        d = VectorOfParameterized(repeat([Normal(0, 1)], 5))
        @test d.distribution == repeat([Normal(0, 1)], 5)
    end
    @testset "ConstraintType" begin
        tol = 10 * eps(Float64)
        # Tests for the ConstraintType
        # The predefined transforms
        c1 = bounded_below(0.2)
        @test isapprox(c1.constrained_to_unconstrained(1.0) - (log(1.0 - 0.2)), 0.0, atol = tol)
        @test isapprox(c1.unconstrained_to_constrained(0.0) - (exp(0.0) + 0.2), 0.0, atol = tol)
        @test get_constraint_type(c1) == BoundedBelow
        @test get_bounds(c1) == Dict("lower_bound" => 0.2)

        c2 = bounded_above(0.2)
        @test isapprox(c2.constrained_to_unconstrained(-1.0) - (-log(0.2 - -1.0)), 0.0, atol = tol)
        @test isapprox(c2.unconstrained_to_constrained(10.0) - (0.2 - exp(-10.0)), 0.0, atol = tol)
        @test get_constraint_type(c2) == BoundedAbove
        @test get_bounds(c2) == Dict("upper_bound" => 0.2)

        c3 = bounded(-0.1, 0.2)
        @test get_bounds(c3) == Dict("lower_bound" => -0.1, "upper_bound" => 0.2)
        @test get_constraint_type(c3) == Bounded
        @test isapprox(c3.constrained_to_unconstrained(0.0) - (log((0.0 - -0.1) / (0.2 - 0.0))), 0.0, atol = tol)
        @test isapprox(
            c3.unconstrained_to_constrained(1.0) - ((0.2 * exp(1.0) + -0.1) / (exp(1.0) + 1)),
            0.0,
            atol = tol,
        )
        @test_throws DomainError bounded(0.2, -0.1)

        #an example with user defined invertible transforms
        c_to_u = (x -> 3 * x + 14)
        jacobian = (x -> 3)
        u_to_c = (x -> (x - 14) / 3)

        abstract type MyConstraint <: ConstraintType end
        c4 = Constraint{MyConstraint}(c_to_u, jacobian, u_to_c, nothing)
        @test isapprox(c4.constrained_to_unconstrained(5.0) - c_to_u(5.0), 0.0, atol = tol)
        @test isapprox(c4.unconstrained_to_constrained(5.0) - u_to_c(5.0), 0.0, atol = tol)
        @test get_constraint_type(c4) == MyConstraint

        #length, size
        @test length(c1) == 1
        @test size(c1) == (1,)

        #equality
        @test c3 == c3
        @test !(c1 == c2)

    end

    @testset "ParameterDistribution: Build and combine" begin

        # create single ParameterDistribution
        d = VectorOfParameterized([Normal(0, 1), Gamma(2.0, 0.8)])
        c_mismatch = no_constraint()
        c_wrongtype = [no_constraint(), 3.0]
        c = [no_constraint(), no_constraint()]
        name = "normal_and_gamma"

        @test_throws ArgumentError ParameterDistribution(d, c_wrongtype, name) #wrong type of constraint
        @test_throws DimensionMismatch ParameterDistribution(d, c_mismatch, name) #wrong number of constraints

        # test checks on stored information
        u = ParameterDistribution(d, c, name)
        @test u.distribution == [d]
        @test u.constraint == c #as c is already a vector
        @test u.name == [name]

        d = Parameterized(Normal(0, 1))
        c = [no_constraint()]
        u = ParameterDistribution(d, c, name)
        @test u.constraint == c #as c is already a vector

        # test concatenation
        d1 = Parameterized(Normal(0, 1))
        c1 = no_constraint()
        name1 = "unconstrained_normal"
        u1 = ParameterDistribution(d1, c1, name1)
        @test u1.constraint == [c1]

        d2 = VectorOfParameterized(repeat([MvNormal(ones(3), 0.2 * I)], 4))
        c2 = repeat([no_constraint()], 12) #3D distribution repeated 4 times has 12 constraints
        name2 = "three_unconstrained_MvNs"
        u2 = ParameterDistribution(d2, c2, name2)

        d3 = Samples([1.0 3.0 5.0 7.0; 9.0 11.0 13.0 15.0])
        c3 = [bounded(10, 15), no_constraint()]
        name3 = "constrained_sampled"
        u3 = ParameterDistribution(d3, c3, name3)

        @test_throws ArgumentError ParameterDistribution([u1, u2])

        u = combine_distributions([u1, u2, u3])
        @test u.distribution == [d1, d2, d3]
        @test u.constraint == cat([[c1], c2, c3]..., dims = 1)
        @test u.name == [name1, name2, name3]

        #equality
        @test u == u
        @test !(u2 == u3)

    end

    @testset "ParameterDistribution: Dictionary interface" begin

        # dictionary - testing test checks on the distribution elements
        d = Parameterized(Normal(0, 1))
        c = no_constraint()
        name = "unconstrained_normal"

        param_dict_fail = [Dict("a" => 1), 1]
        @test_throws ArgumentError ParameterDistribution(param_dict_fail) # not an array of Dicts

        @test_throws ArgumentError ParameterDistribution(param_dict_fail) # not a Dict

        param_dict_fail = Dict("distribution" => d)
        @test_throws ArgumentError ParameterDistribution(param_dict_fail) # not all the keys
        param_dict_fail = Dict("distribution" => 1, "constraint" => c, "name" => name)
        @test_throws ArgumentError ParameterDistribution(param_dict_fail) # not right distribution type
        param_dict_fail = Dict("distribution" => d, "constraint" => 1, "name" => name)
        @test_throws ArgumentError ParameterDistribution(param_dict_fail) # not right constraint type
        param_dict_fail = Dict("distribution" => d, "constraint" => [1, no_constraint(), 3], "name" => name)
        @test_throws ArgumentError ParameterDistribution(param_dict_fail) # not right constraint type
        param_dict_fail = Dict("distribution" => d, "constraint" => c, "name" => 1)
        @test_throws ArgumentError ParameterDistribution(param_dict_fail) # not right name type
        param_dict_fail = Dict("distribution" => d, "constraint" => [no_constraint(), no_constraint()], "name" => name)
        @test_throws DimensionMismatch ParameterDistribution(param_dict_fail) # wrong number of constraints

        param_dict = Dict("distribution" => d, "constraint" => c, "name" => name)
        u = ParameterDistribution(param_dict)
        @test u.distribution == [d]
        @test u.constraint == [c]
        @test u.name == [name]

        c_arr = [no_constraint()]
        param_dict = Dict("distribution" => d, "constraint" => c_arr, "name" => name)
        u = ParameterDistribution(param_dict)
        @test u.constraint == [c]

        # Tests for the ParameterDistribution
        d = Parameterized(MvNormal(zeros(4), 0.1 * I))
        c = [no_constraint(), bounded_below(-1.0), bounded_above(0.4), bounded(-0.1, 0.2)]
        name = "constrained_mvnormal"
        param_dict = Dict("distribution" => d, "constraint" => c, "name" => name)

        u = ParameterDistribution(param_dict)
        @test u.distribution == [d]
        @test u.constraint == c
        @test u.name == [name]
        param_dict_fail = param_dict
        param_dict_fail["constraint"] = c[1]
        @test_throws DimensionMismatch ParameterDistribution(param_dict_fail) # wrong number of constraints

        # Tests for the ParameterDistribution
        d1 = Parameterized(MvNormal(zeros(4), 0.1 * I))
        c1 = [no_constraint(), bounded_below(-1.0), bounded_above(0.4), bounded(-0.1, 0.2)]
        name1 = "constrained_mvnormal"
        param_dict1 = Dict("distribution" => d1, "constraint" => c1, "name" => name1)

        d2 = Samples([1.0 3.0 5.0 7.0; 9.0 11.0 13.0 15.0])
        c2 = [bounded(10, 15), no_constraint()]
        name2 = "constrained_sampled"
        param_dict2 = Dict("distribution" => d2, "constraint" => c2, "name" => name2)

        param_dict_array = [param_dict1, param_dict2]
        u = ParameterDistribution(param_dict_array)
        @test u.distribution == [d1, d2]
        @test u.constraint == cat([c1, c2]..., dims = 1)
        @test u.name == [name1, name2]

    end

    @testset "getter functions" begin
        # setup for the tests:
        d1 = Parameterized(MvNormal(zeros(4), 0.1 * I))
        c1 = [no_constraint(), bounded_below(-1.0), bounded_above(0.4), bounded(-0.1, 0.2)]
        name1 = "constrained_mvnormal"
        u1 = ParameterDistribution(d1, c1, name1)

        d2 = Samples([1 2 3 4])
        c2 = [bounded(10, 15)]
        name2 = "constrained_sampled"
        u2 = ParameterDistribution(d2, c2, name2)

        d3 = VectorOfParameterized(repeat([Beta(2, 2)], 3))
        c3 = repeat([no_constraint()], 3)
        name3 = "vector_beta"
        u3 = ParameterDistribution(d3, c3, name3)

        u = combine_distributions([u1, u2, u3])

        # Test for get_dimension(s)
        @test ndims(u1) == 4
        @test ndims(u2) == 1
        @test ndims(u) == 8

        @test get_dimensions(u1) == [4]
        @test get_dimensions(u2) == [1]
        @test get_dimensions(u) == [4, 1, 3]

        # Tests for get_name
        @test get_name(u1) == [name1]
        @test get_name(u) == [name1, name2, name3]

        # Tests for get_n_samples
        @test typeof(get_n_samples(u)[name1]) <: String
        @test typeof(get_n_samples(u)[name3]) <: String
        @test get_n_samples(u)[name2] == 4

        # Tests for get_distribution
        @test get_distribution(d1) == MvNormal(zeros(4), 0.1 * I)
        @test get_distribution(u1)[name1] == MvNormal(zeros(4), 0.1 * I)
        @test typeof(get_distribution(d2)) == Array{Int64, 2}
        @test typeof(get_distribution(u2)[name2]) == Array{Int64, 2}

        @test get_distribution(d3) == repeat([Beta(2, 2)], 3)
        @test get_distribution(u3)[name3] == repeat([Beta(2, 2)], 3)

        d = get_distribution(u)
        @test d[name1] == MvNormal(zeros(4), 0.1 * I)
        @test typeof(d[name2]) == Array{Int64, 2}

        # Test for get_all_constraints
        @test get_all_constraints(u) == cat([c1, c2, c3]..., dims = 1)
    end

    @testset "statistics functions" begin

        # setup for the tests:
        d1 = Parameterized(MvNormal(zeros(4), 0.1 * I))
        c1 = [no_constraint(), bounded_below(-1.0), bounded_above(0.4), bounded(-0.1, 0.2)]
        name1 = "constrained_mvnormal"

        u1 = ParameterDistribution(d1, c1, name1)

        d2 = Samples([1 2 3 4])
        c2 = [bounded(10, 15)]
        name2 = "constrained_sampled"
        u2 = ParameterDistribution(d2, c2, name2)

        d3 = VectorOfParameterized([Beta(2, 2), MvNormal(zeros(2), 0.1 * I)])
        c3 = [no_constraint(), no_constraint(), no_constraint()] # d3 has 3 dimensions
        name3 = "unconstrained_beta_and_MvN"
        u3 = ParameterDistribution(d3, c3, name3)

        u = combine_distributions([u1, u2])

        d4 = Samples([1 2 3 4 5 6 7 8; 8 7 6 5 4 3 2 1])
        c4 = [no_constraint(), no_constraint()]
        name4 = "constrained_MVsampled"
        u4 = ParameterDistribution(d4, c4, name4)

        v = combine_distributions([u1, u2, u3, u4])

        # Tests for sample distribution
        seed = 2020
        Random.seed!(seed)
        s1 = rand(MvNormal(zeros(4), 0.1 * I), 1)
        Random.seed!(seed)
        @test sample(u1) == s1

        Random.seed!(seed)
        s1 = rand(MvNormal(zeros(4), 0.1 * I), 3)
        Random.seed!(seed)
        @test sample(u1, 3) == s1

        Random.seed!(seed)
        idx = StatsBase.sample(collect(1:size(d2.distribution_samples)[2]), 1)
        s2 = d2.distribution_samples[:, idx]
        Random.seed!(seed)
        @test sample(u2) == s2

        Random.seed!(seed)
        idx = StatsBase.sample(collect(1:size(d2.distribution_samples)[2]), 3)
        s2 = d2.distribution_samples[:, idx]
        Random.seed!(seed)
        @test sample(u2, 3) == s2

        Random.seed!(seed)
        s3 = zeros(3, 1)
        s3[1] = rand(Beta(2, 2))
        s3[2:3] = rand(MvNormal(zeros(2), 0.1 * I))
        Random.seed!(seed)
        @test sample(u3) == s3

        Random.seed!(seed)
        s1 = sample(u1, 3)
        s2 = sample(u2, 3)
        s3 = sample(u3, 3)
        s4 = sample(u4, 3)
        Random.seed!(seed)
        s = sample(v, 3)
        @test s == cat([s1, s2, s3, s4]..., dims = 1)

        #Test for get_logpdf
        @test_throws ErrorException get_logpdf(u, zeros(ndims(u)))
        x_in_bd = [0.5, 0.5, 0.5]
        Random.seed!(seed)
        lpdf3 = sum([logpdf(Beta(2, 2), x_in_bd[1])[1], logpdf(MvNormal(zeros(2), 0.1 * I), x_in_bd[2:3])[1]]) #throws deprecated warning without "."

        Random.seed!(seed)
        @test isapprox(get_logpdf(u3, x_in_bd) - lpdf3, 0.0; atol = 1e-6)
        @test_throws DimensionMismatch get_logpdf(u3, [0.5, 0.5])

        #Test for cov, var        
        block_cov = cat([cov(d1), var(d2), cov(d3), cov(d4)]..., dims = (1, 2))
        @test isapprox(cov(v) - block_cov, zeros(ndims(v), ndims(v)); atol = 1e-6)
        block_var = [block_cov[i, i] for i in 1:size(block_cov)[1]]
        @test isapprox(var(v) - block_var, zeros(ndims(v)); atol = 1e-6)

        #Test for mean
        means = cat([mean(d1), mean(d2), mean(d3), mean(d4)]..., dims = 1)
        @test isapprox(mean(v) - means, zeros(ndims(v)); atol = 1e-6)

    end

    @testset "statistics functions: explict RNG" begin

        # setup for the tests:
        rng_seed = 1234
        test_d = MvNormal(zeros(4), 0.1 * I)
        test_d3a = Beta(2, 2)
        test_d3b = MvNormal(zeros(2), 0.1 * I)
        d0 = Parameterized(test_d)

        d1 = Parameterized(test_d)
        c1 = [no_constraint(), bounded_below(-1.0), bounded_above(0.4), bounded(-0.1, 0.2)]
        name1 = "constrained_mvnormal"
        u1 = ParameterDistribution(d1, c1, name1)

        d2 = Samples([1 2 3 4])
        c2 = [bounded(10, 15)]
        name2 = "constrained_sampled"
        u2 = ParameterDistribution(d2, c2, name2)

        d3 = VectorOfParameterized([Beta(2, 2), MvNormal(zeros(2), 0.1 * I)])
        c3 = repeat([no_constraint()], 3)
        name3 = "beta_and_normal"
        u3 = ParameterDistribution(d3, c3, name3)

        # Tests for sample distribution
        rng1 = Random.MersenneTwister(rng_seed)
        @test sample(copy(rng1), d0) == rand(copy(rng1), test_d, 1)
        @test sample(copy(rng1), d0, 3) == rand(copy(rng1), test_d, 3)

        @test sample(copy(rng1), u1) == rand(copy(rng1), test_d, 1)
        @test sample(copy(rng1), u1, 3) == rand(copy(rng1), test_d, 3)

        idx = StatsBase.sample(copy(rng1), collect(1:size(d2.distribution_samples)[2]), 1)
        s2 = d2.distribution_samples[:, idx]
        @test sample(copy(rng1), u2) == s2
        @test sample(copy(rng1), u2, 1) == s2
        @test sample(copy(rng1), d2) == s2
        @test sample(copy(rng1), d2, 1) == s2

        rng1_new = copy(rng1) # (copy here, as internally u3 samples rng1 sequentially)
        @test sample(copy(rng1), u3) == cat([rand(rng1_new, test_d3a, 1), rand(rng1_new, test_d3b, 1)]..., dims = 1)
        rng1_new = copy(rng1)
        @test sample(copy(rng1), u3, 3) ==
              cat([reshape(rand(rng1_new, test_d3a, 3), :, 3), rand(rng1_new, test_d3b, 3)]..., dims = 1)
        rng1_new = copy(rng1)
        @test sample(copy(rng1), d3) == cat([rand(rng1_new, test_d3a, 1), rand(rng1_new, test_d3b, 1)]..., dims = 1)
        rng1_new = copy(rng1)
        @test sample(copy(rng1), d3, 1) == cat([rand(rng1_new, test_d3a, 1), rand(rng1_new, test_d3b, 1)]..., dims = 1)

        # try it again with different RNG; use StableRNG since Random doesn't provide a 
        # second seedable algorithm on julia <=1.7
        rng2 = StableRNG(rng_seed)
        @test sample(copy(rng2), d0) == rand(copy(rng2), test_d, 1)
        @test sample(copy(rng2), d0, 3) == rand(copy(rng2), test_d, 3)

        @test sample(copy(rng2), u1) == rand(copy(rng2), test_d, 1)
        @test sample(copy(rng2), u1, 3) == rand(copy(rng2), test_d, 3)

        idx = StatsBase.sample(copy(rng2), collect(1:size(d2.distribution_samples)[2]), 1)
        s2 = d2.distribution_samples[:, idx]
        @test sample(copy(rng2), u2) == s2

        rng2_new = copy(rng2)
        @test sample(copy(rng2), u3) == cat([rand(rng2_new, test_d3a, 1), rand(rng2_new, test_d3b, 1)]..., dims = 1)
        rng2_new = copy(rng2)
        @test sample(copy(rng2), u3, 3) ==
              cat([reshape(rand(rng2_new, test_d3a, 3), :, 3), rand(rng2_new, test_d3b, 3)]..., dims = 1)

        # test that optional parameter defaults to Random.GLOBAL_RNG, for all methods.
        # reset the global seed instead of copying the rng object's state
        rng_seed = 2468
        Random.seed!(rng_seed)
        test_lhs = sample(d0)
        Random.seed!(rng_seed)
        @test test_lhs == rand(test_d, 1)

        Random.seed!(rng_seed)
        test_lhs = sample(d0, 3)
        Random.seed!(rng_seed)
        @test test_lhs == rand(test_d, 3)

        Random.seed!(rng_seed)
        test_lhs = sample(u1)
        Random.seed!(rng_seed)
        @test test_lhs == rand(test_d, 1)

        Random.seed!(rng_seed)
        test_lhs = sample(u1, 3)
        Random.seed!(rng_seed)
        @test test_lhs == rand(test_d, 3)

        Random.seed!(rng_seed)
        test_lhs = sample(d0)
        Random.seed!(rng_seed)
        @test test_lhs == rand(test_d, 1)

        Random.seed!(rng_seed)
        idx = StatsBase.sample(collect(1:size(d2.distribution_samples)[2]), 1)
        test_lhs = d2.distribution_samples[:, idx]
        Random.seed!(rng_seed)
        @test test_lhs == sample(u2)

        Random.seed!(rng_seed)
        test_lhs = sample(d3)
        Random.seed!(rng_seed)
        @test test_lhs == cat([rand(test_d3a, 1), rand(test_d3b, 1)]..., dims = 1)

    end

    @testset "transform definitions" begin
        tol = 1e-8
        cons = [
            no_constraint(),
            bounded_below(-1.0),
            bounded_below(2.0),
            bounded_above(-1.0),
            bounded_above(2.0),
            bounded(-1.0, 2.0),
        ]
        for con in cons
            for u0 in [-Inf, -5.0, -1.0, 0.0, 1.0, 2.0, 5.0, Inf]
                c0 = con.unconstrained_to_constrained(u0)
                @test ~isnan(c0)
                u1 = con.constrained_to_unconstrained(c0)
                @test ~isnan(u1)
                if isinf(u0)
                    @test u0 == u1
                else
                    @test isapprox(u0, u1, atol = tol)
                end
            end
        end
        cons = [no_constraint(), bounded_below(-1.0), bounded_above(2.0), bounded(-1.0, 2.0)]
        for con in cons
            for c0 in [-1.0, 0.0, 1.0, 2.0]
                @test ~isnan(con.c_to_u_jacobian(c0))
                @test con.c_to_u_jacobian(c0) >= 0.0
                u0 = con.constrained_to_unconstrained(c0)
                @test ~isnan(u0)
                c1 = con.unconstrained_to_constrained(u0)
                @test ~isnan(c1)
                @test isapprox(c0, c1, atol = tol)
            end
        end

        @testset "transform definitions: handle +/-Inf" begin
            for cons in (bounded(-Inf, Inf), bounded_below(-Inf), bounded_above(Inf))
                @test cons.constrained_to_unconstrained(5.0) == 5.0
                @test cons.unconstrained_to_constrained(5.0) == 5.0
            end
            c_hi = bounded_above(5.0)
            cons = bounded(-Inf, 5.0)
            @test cons.unconstrained_to_constrained(10.0) == c_hi.unconstrained_to_constrained(10.0)
            c_lo = bounded_below(-5.0)
            cons = bounded(-5.0, Inf)
            @test cons.unconstrained_to_constrained(-10.0) == c_lo.unconstrained_to_constrained(-10.0)
        end
    end

    @testset "transform functions" begin
        #setup for the tests
        tol = 1e-8
        d1 = Parameterized(MvNormal(zeros(4), 0.1 * I))
        c1 = [no_constraint(), bounded_below(-1.0), bounded_above(0.4), bounded(-0.1, 0.2)]
        name1 = "constrained_mvnormal"
        param_dict1 = Dict("distribution" => d1, "constraint" => c1, "name" => name1)
        u1 = ParameterDistribution(param_dict1)

        d2 = Samples([1.0 3.0 5.0 7.0; 9.0 11.0 13.0 15.0])
        c2 = [bounded(10, 15), no_constraint()]
        name2 = "constrained_sampled"
        param_dict2 = Dict("distribution" => d2, "constraint" => c2, "name" => name2)
        u2 = ParameterDistribution(param_dict2)

        param_dict = [param_dict1, param_dict2]
        u = ParameterDistribution(param_dict)

        x_unbd = rand(MvNormal(zeros(6), 3 * I), 1000)  #6 x 1000 
        # Tests for transforms
        x_real_constrained1 = mapslices(x -> transform_unconstrained_to_constrained(u1, x), x_unbd[1:4, :]; dims = 1)
        @test isapprox(
            x_unbd[1:4, :] -
            mapslices(x -> transform_constrained_to_unconstrained(u1, x), x_real_constrained1; dims = 1),
            zeros(size(x_unbd[1:4, :]));
            atol = tol,
        )

        x_real_constrained2 = mapslices(x -> transform_unconstrained_to_constrained(u2, x), x_unbd[5:6, :]; dims = 1)
        @test isapprox(
            x_unbd[5:6, :] -
            mapslices(x -> transform_constrained_to_unconstrained(u2, x), x_real_constrained2; dims = 1),
            zeros(size(x_unbd[5:6, :]));
            atol = tol,
        )

        x_real = mapslices(x -> transform_unconstrained_to_constrained(u, x), x_unbd; dims = 1)
        x_unbd_tmp = mapslices(x -> transform_constrained_to_unconstrained(u, x), x_real; dims = 1)
        @test isapprox(x_unbd - x_unbd_tmp, zeros(size(x_unbd)); atol = tol)

        # Tests transforms for other input structures
        @test isapprox(transform_unconstrained_to_constrained(u1, x_unbd[1:4, :]), x_real_constrained1; atol = tol)
        @test isapprox(
            x_unbd[1:4, :] - transform_constrained_to_unconstrained(u1, x_real_constrained1),
            zeros(size(x_unbd[1:4, :]));
            atol = tol,
        )
        @test isapprox(transform_unconstrained_to_constrained(u2, x_unbd[5:6, :]), x_real_constrained2; atol = tol)
        @test isapprox(
            x_unbd[5:6, :] - transform_constrained_to_unconstrained(u2, x_real_constrained2),
            zeros(size(x_unbd[5:6, :]));
            atol = tol,
        )
        @test isapprox(
            transform_unconstrained_to_constrained(u2, [x_unbd[5:6, :], x_unbd[5:6, :]]),
            [x_real_constrained2, x_real_constrained2];
            atol = tol,
        )
    end

    @testset "constrained_gaussian" begin
        @testset "constrained_gaussian: bounds" begin
            @test_throws DomainError constrained_gaussian("test", 0.0, 1.0, 5.0, -5.0)
            @test_throws DomainError constrained_gaussian("test", -10.0, 1.0, -5.0, 5.0)
            @test_throws DomainError constrained_gaussian("test", 10.0, 1.0, -5.0, 5.0)
            # σ too wide relative to constraint interval
            @test_throws DomainError constrained_gaussian("test", 0.0, 10.0, -1.0, 1000.0)
            @test_throws DomainError constrained_gaussian("test", 0.0, 10.0, -1000.0, 1.0)
            @test_throws DomainError constrained_gaussian("test", 0.0, 10.0, -1.0, 1.0)
            # optimization failure
            @test_logs (:warn,) match_mode = :any ParameterDistributions._constrained_gaussian(
                "test",
                0.99999,
                0.000006,
                0.0,
                1.0,
            )
        end
        @testset "constrained_gaussian: closed form" begin
            μ_c = -5.0
            σ_c = 2.0
            pd = constrained_gaussian("test", μ_c, σ_c, -Inf, Inf)
            d = get_distribution(pd)["test"]
            @test mean(d) == μ_c
            @test std(d) == σ_c
            pd = constrained_gaussian("test", μ_c, σ_c, -Inf, 10.0)
            d = get_distribution(pd)["test"]
            μ_u, σ_u = ParameterDistributions._inverse_lognormal_mean_std(10.0 - μ_c, σ_c)
            @test mean(d) == μ_u
            @test std(d) == σ_u
            pd = constrained_gaussian("test", μ_c, σ_c, -20.0, Inf)
            d = get_distribution(pd)["test"]
            μ_u, σ_u = ParameterDistributions._inverse_lognormal_mean_std(μ_c + 20.0, σ_c)
            @test mean(d) == μ_u
            @test std(d) == σ_u

            #multidimension through repeats
            pd_multi = constrained_gaussian("test", μ_c, σ_c, -20.0, Inf, repeats = 10)
            @test ndims(pd_multi) == 10
            @test all(get_distribution(pd_multi)["test"] .== get_distribution(pd)["test"])
            @test all(get_all_constraints(pd_multi) .== pd.constraint)
            @test get_name(pd_multi) == get_name(pd)

            pd_rep0 = constrained_gaussian("test", μ_c, σ_c, -20.0, Inf, repeats = 0)
            @test pd_rep0 == pd
        end

        @testset "constrained_gaussian: integration" begin
            # verify analytic solutions
            μ_0 = -5.0
            σ_0 = 2.0
            μ_c, σ_c = ParameterDistributions._lognormal_mean_std(μ_0, σ_0)
            μ_u, σ_u = ParameterDistributions._inverse_lognormal_mean_std(μ_c, σ_c)
            @test isapprox(μ_0, μ_u, atol = 1e-7, rtol = 1e-7)
            @test isapprox(σ_0, σ_u, atol = 1e-7, rtol = 1e-7)
            μ_0 = 1.0
            σ_0 = 2.0
            μ_u, σ_u = ParameterDistributions._inverse_lognormal_mean_std(μ_0, σ_0)
            μ_c, σ_c = ParameterDistributions._lognormal_mean_std(μ_u, σ_u)
            @test isapprox(μ_0, μ_c, atol = 1e-7, rtol = 1e-7)
            @test isapprox(σ_0, σ_c, atol = 1e-7, rtol = 1e-7)

            # lognormal case
            cons = bounded(0.0, Inf)
            μ, σ = ParameterDistributions._lognormal_mean_std(0.0, 1.0)
            m, s = ParameterDistributions._mean_std(0.0, 1.0, cons)
            @test isapprox(μ, m, atol = 1e-6, rtol = 1e-6)
            @test isapprox(σ, s, atol = 1e-6, rtol = 1e-6)
            μ, σ = ParameterDistributions._lognormal_mean_std(1.0, 2.0)
            m, s = ParameterDistributions._mean_std(1.0, 2.0, cons)
            @test isapprox(μ, m, atol = 1e-6, rtol = 1e-6)
            @test isapprox(σ, s, atol = 1e-6, rtol = 1e-6)

            # logitnormal
            cons = bounded(0.0, 1.0)
            m, s = ParameterDistributions._mean_std(0.0, 1.0, cons)
            @test isapprox(m, 0.5, atol = 1e-5, rtol = 1e-5)
            @test isapprox(s, 0.20827630, atol = 1e-5, rtol = 1e-5)
            m, s = ParameterDistributions._mean_std(1.0, 2.0, cons)
            @test isapprox(m, 0.64772644, atol = 1e-5, rtol = 1e-5)
            @test isapprox(s, 0.29610580, atol = 1e-5, rtol = 1e-5)

        end

        @testset "constrained_gaussian: optimization" begin
            # Analytic form for inital guess
            μ_u, σ_u = ParameterDistributions._constrained_gaussian_guess(0.2, 0.1, 0.0, 1.0)
            @test isapprox(μ_u, -1.54518, atol = 1e-5, rtol = 1e-5)
            @test isapprox(σ_u, 0.69622, atol = 1e-5, rtol = 1e-5)

            # lognormal - analytic
            μ_u = 1.0
            σ_u = 2.0
            μ_c, σ_c = ParameterDistributions._lognormal_mean_std(μ_u, σ_u)
            pd = constrained_gaussian("test", μ_c, σ_c, 0.0, Inf)
            d = pd.distribution[1].distribution
            @test isapprox(μ_u, mean(d), atol = 1e-5, rtol = 1e-5)
            @test isapprox(σ_u, std(d), atol = 1e-5, rtol = 1e-5)

            # logitnormal
            μ_c = 0.2
            σ_c = 0.1
            pd = constrained_gaussian("test", μ_c, σ_c, 0.0, 1.0)
            d = pd.distribution[1].distribution
            c = pd.constraint[1]
            m_c, s_c = ParameterDistributions._mean_std(mean(d), std(d), c)
            @test isapprox(m_c, μ_c, atol = 1e-4, rtol = 1e-3)
            @test isapprox(s_c, σ_c, atol = 1e-4, rtol = 1e-3)
            @test isapprox(-1.5047670627292984, mean(d), atol = 1e-4, rtol = 1e-3)
            @test isapprox(0.6474290829043071, std(d), atol = 1e-4, rtol = 1e-3)
        end
    end
end
